/*----------------------------------------------------------------------------
 Copyright:      Radig Ulrich  mailto: mail@ulrichradig.de
 Author:         Radig Ulrich
 Remarks:        
 known Problems: none
 Version:        24.10.2007
 Description:    Webserver Applikation
----------------------------------------------------------------------------*/

#include "httpd.h"
#include "websites.h"
#include "pictures.h"
#include <stdio.h>

struct http_table http_entry[MAX_TCP_ENTRY];

struct Websites myWebsites;						// enthlt alle Webseiten
struct Pictures myPictures;						// enthlt alle Bilder

//Hier wird das codierte Passwort aus config.h gespeichert.
unsigned char http_auth_passwort[20];

//----------------------------------------------------------------------------
//Variablenarry zum einfgen in Webseite %VA@00 bis %VA@09
unsigned int var_array[MAX_VAR_ARRAY] = {0,0,0,0,0,0,0,2,0,0};
//----------------------------------------------------------------------------

PROGMEM char http_headerText[]={"HTTP/1.0 200 Document follows\r\nServer: AVR_Small_Webserver\r\nContent-Type: text/html\r\n\r\n"};
PROGMEM char http_headerPicture[]={"HTTP/1.0 200 Document follows\r\nServer: AVR_Small_Webserver\r\nContent-Type: image/jpg\r\n\r\n"};

PROGMEM char http_header3[]={	"HTTP/1.0 401 Unauthorized\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"WWW-Authenticate: Basic realm=\"NeedPassword\""
								"\r\nContent-Type: text/html\r\n\r\n"};//LEN 129

//----------------------------------------------------------------------------
//Website fr die Warnmeldung Kein Zugriff Seite bei keinem Passwort
PROGMEM char Page0[] = {"401 Unauthorized%END"};

unsigned char rx_header_end[5] = {"\r\n\r\n\0"};

//----------------------------------------------------------------------------
/// \brief Initialisierung des Httpd Testservers
void httpd_init (void)
{
	//Port fr die schaltbaren Ausgnge
	//Port kann nur bei ENC28J60 benutzt werden 
	//RTL8019 bentigt PORTA
#if USE_ENC28J60
	DDRB |= 0x03; 
#endif

	//mal sehen was frei ist
#if USE_RTL8019
#endif

	//HTTP_AUTH_STRING 
	decode_base64((unsigned char*)HTTP_AUTH_STRING,http_auth_passwort);

	//Serverport und Anwendung eintragen
	add_tcp_app (80, (void(*)(unsigned char))httpd);
	
		 // meine Webseiten initialisieren
	myWebsites.NoSites = 3;

	strcpy(myWebsites.site[0].ShowName, "home\0");
	strcpy(myWebsites.site[0].FileName, "home.html\0");
	myWebsites.site[0].pSite = webHome_html;

	strcpy(myWebsites.site[1].ShowName, "formular\0");
	strcpy(myWebsites.site[1].FileName, "form.html\0");
	myWebsites.site[1].pSite = webForm_html;

	strcpy(myWebsites.site[2].ShowName, "info\0");
	strcpy(myWebsites.site[2].FileName, "info.html\0");
	myWebsites.site[2].pSite = webInfo_html;


	 // meine Bilder initialisieren
	myPictures.NoPictures = 2;

	strcpy(myPictures.picture[0].FileName, "balke.gif\0");
	myPictures.picture[0].pPicture = pctBalken_gif;

//	strcpy(myPictures.picture[1].FileName, "raket.gif\0");
//	myPictures.picture[1].pPicture = pctRackete_gif;

	strcpy(myPictures.picture[1].FileName, "m32.gif\0");
	myPictures.picture[1].pPicture = pctMega32_gif;
	
	strcpy(WordList[0].text, "kalt\0"); WordList[0].length=5;
	strcpy(WordList[1].text, "mittel\0"); WordList[1].length=7;
	strcpy(WordList[2].text, "warm\0"); WordList[2].length=5;
	
}
   
//----------------------------------------------------------------------------
/// \brief Http-Testserverhauptroutine
void httpd (unsigned char index)
{
	//Allererste Aufruf des Ports fr diese Anwendung
	//HTTPD_Anwendungsstack lschen
	if(tcp_entry[index].app_status==1)
	{
		httpd_stack_clear(index);
	}
	
	//HTTP wurde bei dieser Verbindung zum ersten mal aufgerufen oder
	//HTML Header Retransmission!
	if (tcp_entry[index].app_status <= 2)
	{	
		httpd_header_check (index);
	}
	
	//Der Header wurde gesendet und mit ACK besttigt (tcp_entry[index].app_status+1)
	//war das HTML Packet fertig, oder mssen weitere Daten gesendet werden, oder Retransmission?
	if (tcp_entry[index].app_status > 2 && tcp_entry[index].app_status < 0xFFFE)
	{
		httpd_data_send (index);
	}
	
	//Verbindung kann geschlossen werden! Alle HTML Daten wurden gesendet TCP Port kann
	//geschlossen werden (tcp_entry[index].app_status >= 0xFFFE)!!
	if (tcp_entry[index].app_status >= 0xFFFE)
	{
		tcp_entry[index].app_status = 0xFFFE;
		tcp_Port_close(index);
		return;
	}
	
	return;
}

//----------------------------------------------------------------------------
/// \brief Lschen des HTTP-Stacks
void httpd_stack_clear (unsigned char index)
{
		http_entry[index].http_header_type =0;
		http_entry[index].first_switch = 0;
		http_entry[index].http_auth = HTTP_AUTH_DEFAULT;
		http_entry[index].new_page_pointer = 0;
		http_entry[index].old_page_pointer = 0;
		http_entry[index].auth_ptr = http_auth_passwort;
		http_entry[index].hdr_end_pointer = rx_header_end;
		
		http_entry[index].content = 0;
		http_entry[index].header_new_pointer = webHeader_html;
		http_entry[index].header_old_pointer = 0;
}

//----------------------------------------------------------------------------
//Eintreffenden Header vom Client checken
void httpd_header_check (unsigned char index)
{
		unsigned char find = 0;
		
	//finden der Authorization und das Ende im Header auch ber mehrere Packete hinweg!!	
	if(*http_entry[index].hdr_end_pointer != 0)
	{		
		for(unsigned int a=TCP_DATA_START_VAR;a<(TCP_DATA_END_VAR);a++)
		{	
			
			if(!http_entry[index].http_auth) 
			{
				if (eth_buffer[a] != *http_entry[index].auth_ptr++)
				{
					http_entry[index].auth_ptr = http_auth_passwort;
				}
				if(*http_entry[index].auth_ptr == 0) 
				{
					http_entry[index].http_auth = 1;
				}
			}
			
			if (eth_buffer[a] != *http_entry[index].hdr_end_pointer++)
			{
				http_entry[index].hdr_end_pointer = rx_header_end;
			}
			
			//Das Headerende wird mit (CR+LF+CR+LF) angezeigt!
			if(*http_entry[index].hdr_end_pointer == 0) 
			{
				break;
			}
		}
	}

 // nur wenn der Anfang des Headers ausgewertet wird (die nchsten Pakete nicht mehr)	
	if(tcp_entry[index].app_status == 1)
	{	
		 // ****  Herausfinden, welche http-Seite aufgerufen wurde ****
		if (!http_entry[index].http_header_type)
		{
			 // Standardseite setzen, falls keine Seite angegeben wurde (erste seite in myWebsites)
			http_entry[index].new_page_pointer = myWebsites.site[0].pSite;
			http_entry[index].content = 1; // es soll der Header (html-Header und Navigationsleiste zuerst gesendet werden)
			
			 // suche, ob eine spezielle Website angegeben wurde (z.B. home.html oder form.html)
			for(unsigned i = 0; i<myWebsites.NoSites; i++)
			{
				if(strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR], myWebsites.site[i].FileName) !=0)
				{
					http_entry[index].new_page_pointer = myWebsites.site[i].pSite;
					http_entry[index].content = 1; // es soll der Header (html-Header und Navigationsleiste zuerst gesendet werden)
					find = 1;
					break;
				}
			}	
		}
	
		 // **** Herausfinden, ob ein Bild angefordert wurde (nur wenn noch keine html gefunden wurde) ****
//		if(!find)
		{
			 // suche, ob ein Bild angegeben wurde (z.B. test.gif)
			for(unsigned i = 0; i<myPictures.NoPictures; i++)
			{
				if(strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR], myPictures.picture[i].FileName) !=0)
				{
					http_entry[index].http_header_type = HTTP_HEADER_PICTURE;
					http_entry[index].content = 4;
					http_entry[index].new_page_pointer = myPictures.picture[i].pPicture;
					break;
				}
			}	
		}
	}
	
	//Schaltet Ausgnge am AVR je nach Headerinhalt
	//oder Startseite
	if((!http_entry[index].first_switch)&&(http_entry[index].new_page_pointer == webForm_html))
	{
//		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED1=1\0")!=0){ 
//			PORTB ^= (1<<PB0);
//			if(PINB & 0x01) var_array[8] = 0; else var_array[8] = 1;
//			http_entry[index].first_switch = 1;
//		}

		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED1=1\0")!=0)
		{ 
			PORTB |= (1<<PB0);
			http_entry[index].first_switch = 1;
			var_array[8] = 1;
		}

		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED1=0\0")!=0)
		{
			PORTB &= ~(1<<PB0);
			http_entry[index].first_switch = 1;
			var_array[8] = 0;			
		}

//		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED2=2\0")!=0){
//			PORTB ^= (1<<PB1);
//			if(PINB & 0x02) var_array[9] = 0; else var_array[9] = 1;
//			http_entry[index].first_switch = 1;
//		}
		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED2=1\0")!=0)
		{ 
			PORTB |= (1<<PB1);
			http_entry[index].first_switch = 1;
			var_array[9] = 1;
		}

		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED2=0\0")!=0)
		{
			PORTB &= ~(1<<PB1);
			http_entry[index].first_switch = 1;
			var_array[9] = 0;			
		}

	}
	

	//Wurde das Ende vom Header nicht erreicht
	//kommen noch weitere Stcke vom Header!
	if (*http_entry[index].hdr_end_pointer != 0)
	{
		//Der Empfang wird Quitiert und es wird auf weiteres Headerstck gewartet
		tcp_entry[index].status =  ACK_FLAG;
		create_new_tcp_packet(0,index);
		//Warten auf weitere Headerpackete
		tcp_entry[index].app_status = 1;
		return;
	}	
	
	//Wurde das Passwort in den ganzen Headerpacketen gefunden?
	//Wenn nicht dann ausfhren und Passwort anfordern!
	if((!http_entry[index].http_auth) && tcp_entry[index].status&PSH_FLAG)
	{	
		//HTTP_AUTH_Header senden!
		http_entry[index].new_page_pointer = Page0;
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header3,HTTP_AUTH_HEADER_LEN);
		tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
		create_new_tcp_packet(129,index);
		return;
	}
	
	tcp_entry[index].app_status = 2;
	//Seiten Header wird gesendet
	if(http_entry[index].http_header_type == 1)
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_headerPicture,HTTP_OK_HEADER_LEN);
	}
	else
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_headerText,HTTP_OK_HEADER_LEN);
	}
	
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(87,index);
	return;	
}

//----------------------------------------------------------------------------
/// \brief Senden der Daten an den Client
/// \par index Gibt die Nummer des verwendeten HTTP-Streams an
void httpd_data_send (unsigned char index)
{		
	//Passwort wurde im Header nicht gefunden
	if(!http_entry[index].http_auth)
	{
		http_entry[index].new_page_pointer = Page0;
	}
	
	unsigned int a=0;
	unsigned char str_len;
	unsigned char b;
	char var_conversion_buffer[10];
	
	//kein Packet empfangen Retransmission des alten Packetes
	if (tcp_entry[index].status == 0) 
	{
		http_entry[index].new_page_pointer = http_entry[index].old_page_pointer;
	}
	http_entry[index].old_page_pointer = http_entry[index].new_page_pointer;


		 // **** Senden des Seitenkopfes Teil 1 und Teil 3 (Vor und Nach der Navigationsleiste)
		if( (http_entry[index].content == 1) || (http_entry[index].content == 3) )
		{
			for (a = 0;a<(MTU_SIZE-(TCP_DATA_START)-100);a++)
			{
				b = pgm_read_byte(http_entry[index].header_new_pointer++);
				eth_buffer[TCP_DATA_START+a] = b;
		
				 // Ersetzen der Zeichenkette %VA@ durch die Navigationsleiste
				if (strncasecmp_P("%VA@",http_entry[index].header_new_pointer-1,4)==0)
				{
					http_entry[index].header_new_pointer+=3; // Zeiger fr die nchsten Zeichen hinter %VA@ setzen
					http_entry[index].content = 2;
					break;
				}
									 
				 //wurde das Ende des Packetes erreicht?
				 //Verbindung TCP Port kann beim nchsten ACK geschlossen werden
				 //Schleife wird abgebrochen keine Daten mehr!!
				if (strncasecmp_P("%END",http_entry[index].header_new_pointer,4)==0)
				{	
					http_entry[index].content = 4;
					break;
				}
			}

			//Erzeugte Packet kann nun gesendet werden!
			tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
			create_new_tcp_packet(a,index);
			return;
		}
		 // Senden der Navigationsleiste
		else if( http_entry[index].content == 2)
		{

			unsigned int z=0;
			unsigned int z2=0;
			char * zeiger = &eth_buffer[TCP_DATA_START];
			
			for(unsigned char i = 0; i < myWebsites.NoSites; i++)
			{
					if(i == (myWebsites.NoSites-1))
						z = sprintf(zeiger, "<a href=\"%s\"> %s</a>", myWebsites.site[i].FileName, myWebsites.site[i].ShowName);
					else
						z = sprintf(zeiger, "<a href=\"%s\"> %s &middot</a>", myWebsites.site[i].FileName, myWebsites.site[i].ShowName);

					zeiger += z;
					z2+=z;
			}
			http_entry[index].content = 3;
			

			 //Erzeugte Packet kann nun gesendet werden!
			tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
			create_new_tcp_packet(z2,index);
			return;

		}
		 // Senden der ausgewhlten Seite, bzw. eines gewhlten Bildes
		else if(http_entry[index].content == 4)
		{
			for (a = 0;a<(MTU_SIZE-(TCP_DATA_START)-100);a++)
			{
				b = pgm_read_byte(http_entry[index].new_page_pointer++);
				eth_buffer[TCP_DATA_START+a] = b;
			 
				if(strncasecmp_P("%",http_entry[index].new_page_pointer-1,1)==0)
				{
				 	 //Einsetzen der Variablen %VA@00 bis %VA@09 durch den String die Zahl
					if (strncasecmp_P("%VA@",http_entry[index].new_page_pointer-1,4)==0)
					{
						b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
						b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	
						itoa (var_array[b],var_conversion_buffer,10);
						str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);

						memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
						a += str_len-1;
						http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+5;
					}

				 	 //Ersetzen der Variablen %VI@00 bis %VI@09 durch "EIN" oder "AUS"
					if (strncasecmp_P("%VI@",http_entry[index].new_page_pointer-1,4)==0)
					{
						b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
						b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	

						if(var_array[b] !=0 )
							memmove(&var_conversion_buffer, "EIN\0", 4);
						else
							memmove(&var_conversion_buffer, "AUS\0", 4);

						str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);

						memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
						a += str_len-1;
						http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+5;
					}
				 	 //Einsetzen der Variablen %VR@00x bis %VR@09x durch "checked" wenn var_array[00..]=x
					if (strncasecmp_P("%VR@",http_entry[index].new_page_pointer-1,4)==0)
					{
						b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
						b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	

						if(var_array[b] == (pgm_read_byte(http_entry[index].new_page_pointer+5)-48))
							memmove(&var_conversion_buffer, "checked\0", 8);
						else
							memmove(&var_conversion_buffer, "\0", 1);
								
						str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
						memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
						a += str_len-1;
						http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+6;
					}
					
				 	 //Einsetzen der Variablen %VR@00x bis %VR@09x durch "checked" wenn var_array[00..]=x
					if (strncasecmp_P("%VT@",http_entry[index].new_page_pointer-1,4)==0)
					{
						b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
						b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	

						memmove(&var_conversion_buffer, &WordList[var_array[b]].text, WordList[var_array[b]].length);
//printf("\n%s\n%s", WordList[var_array[b]].text, var_conversion_buffer);
						str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
						memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
						a += str_len-1;
						http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+5;
					}
				}
			
				 //wurde das Ende des Paketes erreicht?
				 //Verbindung TCP Port kann beim nchsten ACK geschlossen werden
				 //Schleife wird abgebrochen keine Daten mehr!!
				if (strncasecmp_P("%END",http_entry[index].new_page_pointer,4)==0)
				{	
					tcp_entry[index].app_status = 0xFFFD;
					break;
				}
			}
	}
	//Erzeugte Packet kann nun gesendet werden!
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(a,index);
	return;
}


